﻿using UnityEngine;
using System.Threading;
using System.Collections.Generic;
using System;
using HarmonyLib;
using Steamworks;

public class VoiceChat : Mod
{
    static string activeMic;
    static Thread micThread;
    static List<string> micData;
    public static Messages messageID = (Messages)1234;
    static int channelID = 1235;
    static bool openMic;
    static string pushToTalk = "";
    static bool listen
    {
        get { return openMic || (ExtraSettingsAPI_Loaded && MyInput.GetButton(pushToTalk)); }
    }
    static bool last;
    Harmony harmony;
    static List<CSteamID> connections;
    static List<AudioSource> players;
    static List<float> lastRec;
    public void Start()
    {
        activeMic = "";
        openMic = true;
        last = false;
        micData = new List<string>();
        players = new List<AudioSource>();
        connections = new List<CSteamID>();
        lastRec = new List<float>();
        if (Microphone.devices.Length == 0)
            Debug.LogError("No microphones found!");
        else
        {
            activeMic = Microphone.devices[0];
            Debug.Log("Active microphone has defaulted to: " + activeMic);
        }
        if (RAPI.GetLocalPlayer() != null)
            CreatePlayers();
        Semih_Network.OnPlayerInitialized += CreatePlayer;
        harmony = new Harmony("com.aidanamite.VoiceChat");
        harmony.PatchAll();
        Log("Mod has been loaded!");
    }

    public void Update()
    {
        if (listen != last)
        {
            if (last)
                StopListening();
            else
                StartListening();
            last = !last;
        }
        while (micData.Count > 0)
        {
            Debug.Log(micData[0]);
            micData.RemoveAt(0);
        }
        NetworkMessage m = RAPI.ListenForNetworkMessagesOnChannel(channelID);
        while (m != null && m.message != null)
        {
            if (m.message.Type == messageID)
            {
                Message_VoiceData msg = m.message as Message_VoiceData;
                int ind = connections.IndexOf(m.steamid);
                if (ind != -1)
                {
                    AudioSource player = players[ind];
                    AudioClip clip;
                    if (player.clip == null || player.clip.frequency != msg.frequency)
                    {
                        clip = AudioClip.Create("voice", 200000, 1, msg.frequency, false);
                        player.clip = clip;
                        player.Play();
                    }
                    else
                    {
                        clip = player.clip;
                    }
                    clip.SetData(msg.data, msg.pos);
                    player.time = msg.pos / 200000f;
                    lastRec[ind] = 0;
                }
            }
            m = RAPI.ListenForNetworkMessagesOnChannel(channelID);
        }
        float timeDif = Time.deltaTime;
        for (int i = 0; i < players.Count; i++)
        {
            if ((lastRec[i] += timeDif) > 0.25 && players[i].clip != null)
                players[i].clip.SetData(new float[200000],0);
        }
    }

    

    public void OnModUnload()
    {
        StopListening();
        foreach (AudioSource source in players)
            Destroy(source.gameObject);
        harmony.UnpatchAll(harmony.Id);
        Log("Mod has been unloaded!");
    }

    public override void WorldEvent_WorldUnloaded()
    {
        players.Clear();
        connections.Clear();
        lastRec.Clear();
    }
    public override void WorldEvent_OnPlayerDisconnected(CSteamID steamid, DisconnectReason disconnectReason)
    {
        int ind = connections.IndexOf(steamid);
        players.RemoveAt(ind);
        connections.RemoveAt(ind);
        lastRec.RemoveAt(ind);
    }

    public static void CreatePlayers()
    {
        foreach (KeyValuePair<CSteamID, Network_Player> pair in ComponentManager<Semih_Network>.Value.remoteUsers)
        {
            if (pair.Value.IsLocalPlayer) continue;
            connections.Add(pair.Key);
            lastRec.Add(0);
            CreatePlayer(pair.Value.transform);
        }
    }

    public static void CreatePlayer(Network_Player player)
    {
        if (player.IsLocalPlayer) return;
        lastRec.Add(0);
        connections.Add(ComponentManager<Semih_Network>.Value.GetIDFromPlayer(player));
        CreatePlayer(player.transform);
    }

    public static void CreatePlayer(Transform parent)
    {
        GameObject gO = new GameObject("Voice");
        gO.transform.SetParent(parent, false);
        AudioSource player = gO.AddComponent<AudioSource>();
        player.loop = true;
        player.maxDistance = 30;
        player.minDistance = 0;
        player.volume = 2f;
        player.rolloffMode = AudioRolloffMode.Linear;
        player.spatialBlend = 1;
        players.Add(player);
    }

    static void StartListening()
    {
        if (micThread != null && micThread.IsAlive)
        {
            Debug.LogWarning("Already Listening");
            return;
        }
        if (RAPI.GetLocalPlayer() == null)
        {
            Debug.LogWarning("Not Connected");
            return;
        }
        micThread = new Thread(new ParameterizedThreadStart( delegate
        {
            micData.Add("Started Listening");
            float[] tmp2 = new float[1];
            int lastPos = 0;
            Microphone.GetDeviceCaps(activeMic, out int min, out int max);
            AudioClip tmp = Microphone.Start(activeMic, true, 1, max);
            while (Microphone.IsRecording(activeMic))
            {
                try
                {
                    Thread.Sleep(50);
                    int pos = Microphone.GetPosition(activeMic);
                    tmp2 = new float[pos - lastPos + ((pos < lastPos) ? 200000 : 0)];
                    tmp.GetData(tmp2, lastPos);
                    if (ComponentManager<Semih_Network>.Value != null && ComponentManager<Semih_Network>.Value.HostID.m_SteamID != 0)
                        RAPI.SendNetworkMessage(new Message_VoiceData(tmp2, max, lastPos), channelID);
                    lastPos = pos;
                } catch (Exception err)
                {
                    micData.Add("An error occured in the recording loop\n" + err.GetType().Name + ": " + err.Message + "\n" + err.StackTrace);
                }
            }
            micData.Add("Stopped Listening");
            micThread.Abort();
        }));
        micThread.Start(micThread);
    }

    public static void StopListening()
    {
        Microphone.End(activeMic);
    }

    public void ExtraSettingsAPI_Load()
    {
        pushToTalk = ExtraSettingsAPI_GetKeybindName("talk");
        openMic = ExtraSettingsAPI_GetCheckboxState("open");
    }

    public void ExtraSettingsAPI_SettingsClose()
    {
        openMic = ExtraSettingsAPI_GetCheckboxState("open");
    }

    static Traverse ExtraSettingsAPI_Traverse;
    static bool ExtraSettingsAPI_Loaded = false;
    public bool ExtraSettingsAPI_GetCheckboxState(string SettingName)
    {
        if (ExtraSettingsAPI_Loaded)
            return ExtraSettingsAPI_Traverse.Method("getCheckboxState", new object[] { this, SettingName }).GetValue<bool>();
        return false;
    }
    public string ExtraSettingsAPI_GetKeybindName(string SettingName)
    {
        if (ExtraSettingsAPI_Loaded)
            return ExtraSettingsAPI_Traverse.Method("getKeybindName", new object[] { this, SettingName }).GetValue<string>();
        return "";
    }
}

[Serializable]
public class Message_VoiceData : Message
{
    public float[] data;
    public int frequency;
    public int pos;
    public Message_VoiceData(float[] Data, int Max, int Position) : base(VoiceChat.messageID)
    {
        data = Data;
        frequency = Max;
        pos = Position;
    }
}

[HarmonyPatch(typeof(Semih_Network),"LeaveGame")]
static class Patch_LeaveGame
{
    static void Postfix()
    {
        VoiceChat.StopListening();
    }
}